/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>

#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <ncurses.h>

#include "monitor.h"

#include "stdio.h"
#define  DEBUG "/root/work/topvhost.txt"
#define  WBUF_LEN 128

using namespace std;
/**
 * Local control structure
 */
static topvhost app;
/**
 * Handle KILL and INT signals
 */
static void catchsig(int s)
{
   app.exit = s;
}
/**
 * Handle window resize signal
 */
static void catchwinch(int s)
{
   app.resize = true;
}
/**
 * Initalize the application - called from monitor completion
 */
void topvhost::begin(appOptions options)
{
   WINDOW *win = initscr();
   signal(SIGINT, &catchsig);
   signal(SIGKILL, &catchsig);
   signal(SIGWINCH, &catchwinch);
   this->paint = true;
   this->window  = win;
   this->wyx();
   cbreak();
   nodelay(win,true);
}
/**
 * Diagnostic to display field associations
 */
void topvhost::diagMap(unsigned nidx)
{
   map<int,int>::iterator idb;
   string sep="";
   unsigned j;

   cout << "Output (Needs): '" << this->rfmt << "'" << endl;
   for(j=0;j<this->needs.size();j++) {
      cout << sep << j << " = " << this->needs[j];
      sep = ", ";
      }
   cout << endl << "Input (Gets): '" << this->logfmt << "'" << endl;
   sep = "";
   for(j=0;j<this->gets.size();j++) {
      cout << sep << j << " = " << this->gets[j];
      sep = ", ";
      }
   cout << endl << "Transfer Map (Gets => Needs)" << endl;
   sep = "";
   for(idb = this->xfer.begin();idb != this->xfer.end();++idb) {
      cout << sep << idb->first << " => " << idb->second;
      sep = ", ";
      }
   cout << endl << "Domain name @" << nidx << endl;
}
/**
 * Shutdown the application
 */
void topvhost::end()
{
   if (this->window) {
      endwin();
      this->window = 0;
      }
}
/**
 * Handle the keyboard
 */
int topvhost::input()
{
   int ch;

   if (this->exit || !this->window) return 'q';
   if (this->resize) {
      this->resize = 0;
      endwin();
      refresh();
      this->wyx();
      }
   switch(ch=getch()) {
      case 'd':
         this->debug = !this->debug;
         break;
      case 'h':   case 'H':
         this->order = topvhost::BY_HITS;
         this->paint = true;
         break;
      case 'n':   case 'N':
         this->order = topvhost::BY_NAME;
         this->paint = true;
         break;
      case 'q':   case 'Q':
         ch = 'q';
         break;
      case 'r':   case 'R':
         ch = 'r';
         break;
      case 't':   case 'T':
         this->order = topvhost::BY_ACCESS;
         this->paint = true;
         break;
      }
   return ch;
}
/**
 * Test format character
 */
bool topvhost::isfmt(int c, bool first)
{
   switch(c) {
      case '-':   case '+':   case '#':   case ' ':
         return first;
      case '0':
         return true;
      case '1':   case '2':   case '3':   case '4':   case '5':
      case '6':   case '7':   case '8':   case '9':   case '.':
         return first? false : true;
      }
   return false;
}
/**
 * Thin wrapper for curses macro
 */
void topvhost::wmove(int y,int x)
{
   move(y,x);
}
/**
 * Thin wrappers for curses
 */
void topvhost::wprint(string fmt, float val)
{
   char buf[WBUF_LEN];
   fmt += 'g';
   snprintf(buf,WBUF_LEN,fmt.c_str(), val);
   this->wputs(buf);
}
/**
 * Thin wrappers for curses
 */
void topvhost::wprint(string fmt, int val)
{
   char buf[WBUF_LEN];
   fmt += "d";
   snprintf(buf,WBUF_LEN,fmt.c_str(), val);
   this->wputs(buf);
}
/**
 * Thin wrappers for curses
 */
void topvhost::wprint(string fmt, long val)
{
   char buf[WBUF_LEN];
   fmt += "ld";
   snprintf(buf,WBUF_LEN,fmt.c_str(), val);
   this->wputs(buf);
}
/**
 * Thin wrapper for curses macro
 */
void topvhost::wput(int ch)
{
   if (ch=='\n'||this->wx()<(this->maxx-1))
      addch(ch);
}
/**
 * Thin wrapper for curses macro
 */
void topvhost::wputs(const char *str)
{
   while(*str)
      this->wput(*str++);
}
/**
 * Thin wrapper for curses macro
 */
int topvhost::wx()
{
   return getcurx(static_cast<WINDOW *>(this->window));
}
/**
 * Thin wrapper for curses macro
 */
int topvhost::wy()
{
   return getcury(static_cast<WINDOW *>(this->window));
}
/**
 * Get screen size
 */
void topvhost::wyx()
{
   WINDOW*win = static_cast<WINDOW *>(this->window);
   getmaxyx(win,this->maxy,this->maxx);
}
/**
 * Main loop
 */
int main(int argc, char **argv)
{
   static char * cmds[] = {
      "d", "init",   "0", "Enable diagnostics",
      "f", "file",   "1", "Configuration file",
      "i", "init",   "0", "Parse logs at startup",
      "s", "select", "1", "Select Configuration",
      "v", "verify", "0", "Verify Configuration",
      "h", "help",   "0", "This help"
      };
   static struct appOptions options = {"", "", false, false };
   int nopts = sizeof(cmds)/(4*sizeof(char *));
   struct option* long_options = new option[nopts];
   string short_options;
   int c,i,j,x0,y0;
   Monitor mon;

   for(i=j=0;i<nopts;i++,j+=4) {
      long_options[i].name      = cmds[j+1];
      long_options[i].has_arg   = atoi(cmds[j+2]);
      long_options[i].flag      = NULL;
      long_options[i].val       = cmds[j][0];
      short_options += cmds[j];
      if (long_options[i].has_arg!=0) short_options += ":";
      }
   do {
      c = getopt_long (argc, argv, short_options.c_str(), long_options, NULL);
      switch(c) {
         case 'd':
            app.debug = 1;
            break;
         case 'f':
            options.configFile = optarg;
            break;
         case 'i':
            options.init   = true;
            break;
         case 's':
            options.configSection = optarg;
            break;
         case 'h':
            cerr << "\nUsage: " << argv[0] << " [options]" << endl << endl;
            cerr << "Monitor Virtual Hosts" << endl << endl;
            cerr << "  " << "Options:" << endl;
            for(i=0;long_options[i].name!=NULL;i++) {
               char buf[128];
               sprintf(buf,"--%-10s, -%c %s %s",
                  long_options[i].name, long_options[i].val,
                  long_options[i].has_arg? "[]":"  ",cmds[i*4+3]);
               cerr << "     " << buf << endl;
               }
            cerr << endl;
            exit(1);
         case 'v':
            options.verify = true;
         }
      } while (c!=-1);

   try {
      mon.init(&app, options);
      while(mon.getInput(&app)) {
         vhost *inst = mon.getHosts(&app);
         mon.paintHeader(&app);
         y0 = app.wy();
         x0 = app.wx();
         printw("\n");
         for(i=0;i<app.total;i++) {
            inst[i].paintHost(&app);
            if (app.wy()>=(app.maxy-1)) break;
            }
         move(y0,x0);
         refresh();
         sleep(1);
         }
      mon.quit(&app);
      if (app.exit)
         cerr << " Caught signal " << app.exit << endl;
      }
   catch (exception &e) {
      mon.quit(&app);
      cerr << e.what() << endl;
      }
   return 0;
}
